static const pop_entry pop_d[] =
{ // Dungeon (OOD cap: 31)
  { -3,  7,  514, SEMI, MONS_GIANT_NEWT },
  {  1,  7,  999, FLAT, MONS_GOBLIN },
  { -2,  8,  192, PEAK, MONS_BALL_PYTHON },
  { -1,  7,  136, SEMI, MONS_OOZE },
  { -2,  8,  514, SEMI, MONS_GIANT_COCKROACH },
  {  0, 10,  999, FLAT, MONS_BAT },
  {  0, 10,  999, FLAT, MONS_RAT },
  {  0, 10,  999, FLAT, MONS_KOBOLD },
  {  1, 11,  514, SEMI, MONS_GIANT_GECKO },
  {  1, 11,  334, SEMI, MONS_JACKAL },
  {  1, 11,  192, PEAK, MONS_GIANT_MITE },
  {  1, 11,  999, FLAT, MONS_HOBGOBLIN },
  {  1, 11,  192, PEAK, MONS_GNOLL },
  {  2, 12,  923, FLAT, MONS_ADDER },
  {  2, 12,  514, SEMI, MONS_QUOKKA },
  {  2, 12,  514, SEMI, MONS_WORM },
  {  4, 10,   52, SEMI, MONS_GNOLL_SHAMAN },
  {  3, 13,  999, FLAT, MONS_ORC },
  {  4, 12,  136, SEMI, MONS_ORC_PRIEST },
  {  4, 14,  514, SEMI, MONS_IGUANA },
  {  4, 14,  826, SEMI, MONS_HOUND },
  {  4, 14,  334, SEMI, MONS_GIANT_EYEBALL },
  {  4, 14,  923, FLAT, MONS_WORKER_ANT },
  {  4, 14,  514, SEMI, MONS_SCORPION },
  {  5, 13,   89, PEAK, MONS_FUNGUS },
  {  5, 13,   89, PEAK, MONS_GNOLL_SERGEANT },
  {  4, 14,  514, SEMI, MONS_ORC_WIZARD },
  {  4, 14,  826, SEMI, MONS_OGRE },
  {  4, 14,  334, SEMI, MONS_PHANTOM },
  {  4, 14,  999, FLAT, MONS_SKELETON },
  {  5, 15,  192, PEAK, MONS_WATER_MOCCASIN },
  {  5, 15,  514, SEMI, MONS_JELLY },
  {  5, 15,  514, SEMI, MONS_ICE_BEAST },
  {  5, 15,  514, SEMI, MONS_SKY_BEAST },
  {  5, 15,  826, SEMI, MONS_CENTAUR },
  {  5, 15,  999, FLAT, MONS_CRIMSON_IMP },
  {  5, 15,  192, PEAK, MONS_QUASIT },
  {  5, 15,  192, PEAK, MONS_NECROPHAGE },
  {  5, 15,  999, FLAT, MONS_ZOMBIE },
  {  6, 16,  999, FLAT, MONS_GIANT_FROG },
  {  7, 15,   89, PEAK, MONS_STEAM_DRAGON },
  {  6, 16,  514, SEMI, MONS_GOLIATH_BEETLE },
  {  6, 16,  334, SEMI, MONS_GIANT_SPORE },
  {  6, 16,  826, SEMI, MONS_ORC_WARRIOR },
  {  6, 13,  220, SEMI, MONS_BLACK_BEAR },
  {  6, 16,  192, PEAK, MONS_MUMMY },
  {  6, 16,  334, SEMI, MONS_WIGHT },
  {  7, 17,  192, PEAK, MONS_CROCODILE },
  {  7, 17,  514, SEMI, MONS_HIPPOGRIFF },
  {  7, 17,  514, SEMI, MONS_KILLER_BEE },
  {  7, 17,  192, PEAK, MONS_YELLOW_WASP },
  {  7, 17,  350, SEMI, MONS_HUNGRY_GHOST },
  {  8, 18,  420, SEMI, MONS_MANTICORE },
  {  8, 18,  334, SEMI, MONS_WYVERN },
  {  8, 18,  230, PEAK, MONS_EYE_OF_DRAINING },
  {  9, 17,  136, SEMI, MONS_PLANT },
  {  9, 19,  826, SEMI, MONS_YAK },
  {  9, 19,  334, SEMI, MONS_GIANT_CENTIPEDE },
  { 11, 17,   25, PEAK, MONS_OKLOB_PLANT },
  {  9, 19,  514, SEMI, MONS_BIG_KOBOLD },
  {  9, 19,  334, SEMI, MONS_TWO_HEADED_OGRE },
  {  9, 19,  826, SEMI, MONS_TROLL },
  { 10, 20,  192, PEAK, MONS_KOMODO_DRAGON },
  { 10, 20,  258, PEAK, MONS_SOLDIER_ANT },
  { 10, 20,  230, PEAK, MONS_TRAPDOOR_SPIDER },
  { 10, 20,  999, FLAT, MONS_HILL_GIANT },
  { 10, 20,  514, SEMI, MONS_UNSEEN_HORROR },
  { 10, 20,  334, SEMI, MONS_WRAITH },
  { 11, 21,  350, SEMI, MONS_SHADOW },
  { 11, 21,  258, PEAK, MONS_BASILISK },
  { 11, 21,  514, SEMI, MONS_HYDRA },
  { 13, 19,   25, PEAK, MONS_CATOBLEPAS },
  { 11, 21,  146, PEAK, MONS_BRAIN_WORM },
  { 11, 21,  334, SEMI, MONS_MOTTLED_DRAGON },
  { 13, 19,   52, SEMI, MONS_EYE_OF_DEVASTATION },
  { 11, 21,  826, SEMI, MONS_CYCLOPS },
  { 13, 19,   52, SEMI, MONS_EFREET },
  { 11, 21,  192, PEAK, MONS_SKELETAL_WARRIOR },
  { 14, 20,   52, SEMI, MONS_BLINK_FROG },
  { 12, 22,  879, SEMI, MONS_SLIME_CREATURE },
  { 12, 22,  514, SEMI, MONS_VAMPIRE_MOSQUITO },
  { 14, 20,   25, PEAK, MONS_STONE_GOLEM },
  { 14, 20,   25, PEAK, MONS_WANDERING_MUSHROOM },
  { 16, 18,    4, SEMI, MONS_GUARDIAN_SERPENT },
  { 13, 21,   89, PEAK, MONS_RAKSHASA },
  { 12, 22,  192, PEAK, MONS_VAMPIRE },
  { 12, 22,  999, FLAT, MONS_ZOMBIE },
  { 13, 23,  334, SEMI, MONS_GRIFFON },
  { 13, 23,  230, PEAK, MONS_ROCK_WORM },
  { 13, 23,  999, FLAT, MONS_UGLY_THING },
  { 15, 21,   52, SEMI, MONS_CLAY_GOLEM },
  { 14, 22,  136, SEMI, MONS_BOGGART },
  { 13, 23,  923, FLAT, MONS_CENTAUR_WARRIOR },
  { 14, 22,   89, PEAK, MONS_WIZARD },
  { 13, 23,  674, SEMI, MONS_SHAPESHIFTER },
  { 14, 22,  272, SEMI, MONS_SIMULACRUM },
  { 14, 24,  192, PEAK, MONS_DRAGON },
  { 14, 24,  258, PEAK, MONS_GIANT_AMOEBA },
  { 17, 21,   11, PEAK, MONS_GARGOYLE },
  { 16, 22,   40, PEAK, MONS_KOBOLD_DEMONOLOGIST },
  { 14, 24,  999, FLAT, MONS_SKELETON },
  { 15, 25,  217, PEAK, MONS_GIANT_SLUG },
  { 15, 25,  273, PEAK, MONS_WOLF_SPIDER },
  { 17, 23,   25, PEAK, MONS_IRON_GOLEM },
  { 19, 21,    3, SEMI, MONS_TOENAIL_GOLEM },
  { 15, 25,  334, SEMI, MONS_YAKTAUR },
  { 15, 25,  420, SEMI, MONS_OGRE_MAGE },
  { 17, 25,   89, PEAK, MONS_AGATE_SNAIL },
  { 16, 26,  192, PEAK, MONS_SPINY_WORM },
  { 17, 25,   89, PEAK, MONS_ICE_DRAGON },
  { 17, 25,  136, SEMI, MONS_GREAT_ORB_OF_EYES },
  { 18, 24,   25, PEAK, MONS_GIANT_ORANGE_BRAIN },
  { 17, 25,   89, PEAK, MONS_HARPY },
  { 17, 25,   89, PEAK, MONS_FIRE_CRAB },
  { 17, 25,  136, SEMI, MONS_RED_WASP },
  { 17, 25,   89, PEAK, MONS_VERY_UGLY_THING },
  { 20, 22,    6, FLAT, MONS_EARTH_ELEMENTAL },
  { 20, 22,    6, FLAT, MONS_FIRE_ELEMENTAL },
  { 20, 22,    6, FLAT, MONS_AIR_ELEMENTAL },
  { 16, 26,  690, SEMI, MONS_ORC_KNIGHT },
  { 17, 25,  136, SEMI, MONS_DEEP_ELF_FIGHTER },
  { 17, 25,  136, SEMI, MONS_DEEP_ELF_KNIGHT },
  { 19, 23,    8, PEAK, MONS_DEEP_ELF_MAGE },
  { 19, 23,    8, PEAK, MONS_DEEP_ELF_SUMMONER },
  { 19, 23,    8, PEAK, MONS_NAGA_WARRIOR },
  { 19, 23,    8, PEAK, MONS_NAGA_MAGE },
  { 18, 24,   52, SEMI, MONS_NECROMANCER },
  { 18, 24,   25, PEAK, MONS_SHADOW_WRAITH },
  { 21, 23,    3, SEMI, MONS_PULSATING_LUMP },
  { 17, 27,  244, PEAK, MONS_BOULDER_BEETLE },
  { 19, 25,   35, PEAK, MONS_ORC_HIGH_PRIEST },
  { 19, 27,   66, PEAK, MONS_BORING_BEETLE },
  { 19, 27,   89, PEAK, MONS_REDBACK },
  { 19, 27,   89, PEAK, MONS_SPHINX },
  { 21, 25,    8, PEAK, MONS_CRYSTAL_GOLEM },
  { 21, 25,    8, PEAK, MONS_VAPOUR },
  { 18, 28,  180, PEAK, MONS_FLAYED_GHOST },
  { 18, 28,  258, PEAK, MONS_FREEZING_WRAITH },
  { 23, 25,    3, SEMI, MONS_SHINING_EYE },
  { 21, 27,   25, PEAK, MONS_ORC_SORCERER },
  { 23, 27,   14, SEMI, MONS_ORC_WARLORD },
  { 20, 30,  826, SEMI, MONS_YAKTAUR_CAPTAIN },
  { 21, 29,  136, SEMI, MONS_IRON_TROLL },
  { 24, 28,    8, PEAK, MONS_DANCING_WEAPON },
  { 22, 30,  136, SEMI, MONS_DEEP_TROLL },
  { 21, 31,  192, PEAK, MONS_ETTIN },
  { 21, 31,  571, SEMI, MONS_STONE_GIANT },
  { 22, 30,  136, SEMI, MONS_FIRE_GIANT },
  { 22, 30,  136, SEMI, MONS_FROST_GIANT },
  { 22, 30,  136, SEMI, MONS_DEEP_TROLL_EARTH_MAGE },
  { 22, 30,  136, SEMI, MONS_DEEP_TROLL_SHAMAN },
  { 24, 30,  215, SEMI, MONS_TENGU_REAVER },
  { 23, 29,   25, PEAK, MONS_HELL_KNIGHT },
  { 22, 30,   66, PEAK, MONS_LICH },
  { 23, 31,   89, PEAK, MONS_STORM_DRAGON },
  { 26, 28,    6, FLAT, MONS_DEEP_ELF_CONJURER },
  { 23, 31,   89, PEAK, MONS_PHANTASMAL_WARRIOR },
  { 27, 29,    4, SEMI, MONS_DEEP_ELF_PRIEST },
  { 23, 33,  258, PEAK, MONS_GLOWING_SHAPESHIFTER },
  { 24, 32,   66, PEAK, MONS_TENTACLED_MONSTROSITY },
  { 27, 35,   89, PEAK, MONS_SHADOW_DRAGON },
  { 29, 33,    8, PEAK, MONS_BONE_DRAGON },
  { 29, 33,    8, PEAK, MONS_QUICKSILVER_DRAGON },
  { 29, 33,    8, PEAK, MONS_IRON_DRAGON },
  { 29, 33,   14, SEMI, MONS_GOLDEN_DRAGON },
  { 29, 31,    8, UP,   MONS_DEEP_ELF_HIGH_PRIEST },
  { 30, 31,    4, UP,   MONS_DEEP_ELF_DEMONOLOGIST },
  { 30, 31,    4, UP,   MONS_DEEP_ELF_ANNIHILATOR },
  { 30, 31,    4, UP,   MONS_DEEP_ELF_SORCERER },
  { 30, 31,    4, UP,   MONS_DEEP_ELF_DEATH_MAGE },
  { 28, 34,   25, PEAK, MONS_TITAN },
  { 30, 31,    4, UP,   MONS_ANCIENT_LICH },
  { 0,0,0,FLAT,MONS_0 }
};

static const pop_entry pop_temple[] =
{ // Temple
  { 0,0,0,FLAT,MONS_0 }
};

static const pop_entry pop_orc[] =
{ // Orcish Mines
  {  1,  4,   25, DOWN, MONS_FUNGUS },
  {  1,  4,  192, FLAT, MONS_GOBLIN },
  {  1,  4, 1000, FLAT, MONS_ORC },
  {  1,  4,   89, FLAT, MONS_HOBGOBLIN },
  {  1,  4,  192, FLAT, MONS_ORC_WARRIOR },
  {  1,  4,   25, SEMI, MONS_ORC_PRIEST },
  {  1,  4,   40, FLAT, MONS_WARG },
  {  2,  4,    3, SEMI, MONS_GNOLL },
  {  1,  5,   25, SEMI, MONS_ORC_WIZARD },
  {  1,  5,   25, SEMI, MONS_ORC_KNIGHT },
  {  1,  4,   89, FLAT, MONS_OGRE },
  {  4,  4,    1, FLAT, MONS_GNOLL_SHAMAN },
  {  4,  4,    1, FLAT, MONS_GNOLL_SERGEANT },
  {  2,  4,    8, UP,   MONS_ORC_HIGH_PRIEST },
  {  1,  4,   25, UP,   MONS_ORC_SORCERER },
  {  2,  4,    8, UP,   MONS_ORC_WARLORD },
  {  2,  4,    8, UP,   MONS_TWO_HEADED_OGRE },
  {  4,  4,    1, FLAT, MONS_OGRE_MAGE },
  {  1,  4,   40, UP,   MONS_TROLL },
  {  3,  4,    4, UP,   MONS_IRON_TROLL },
  {  2,  4,    8, UP,   MONS_CYCLOPS },
  {  3,  4,    4, UP,   MONS_ETTIN },
  {  3,  4,    4, UP,   MONS_STONE_GIANT },
  { 0,0,0,FLAT,MONS_0 }
};

static const pop_entry pop_elf[] =
{ // Elven Halls (OOD cap: 7)
  {  1,  7, 1000, FLAT, MONS_FUNGUS },
  {  1,  7, 1000, FLAT, MONS_DEEP_ELF_FIGHTER },
  {  1,  7, 1200, FLAT, MONS_DEEP_ELF_MAGE },
  {  1,  7,  650, FLAT, MONS_DEEP_ELF_SUMMONER },
  {  1,  7,  780, FLAT, MONS_DEEP_ELF_CONJURER },
  {  1,  6,  150, FLAT, MONS_SHAPESHIFTER },
  {  1,  5,   50, FLAT, MONS_ORC_HIGH_PRIEST },
  {  1,  6,   50, FLAT, MONS_ORC_SORCERER },
  {  1,  7,  925, FLAT, MONS_DEEP_ELF_KNIGHT },
  {  1,  7,  350, FLAT, MONS_DEEP_ELF_PRIEST },
  {  1,  6,   65, SEMI, MONS_GLOWING_SHAPESHIFTER },
  {  1,  7,   40, SEMI, MONS_DEEP_ELF_HIGH_PRIEST },
  {  1,  8,   65, SEMI, MONS_DEEP_ELF_DEMONOLOGIST },
  {  1,  7,   40, SEMI, MONS_DEEP_ELF_ANNIHILATOR },
  {  1,  8,   65, SEMI, MONS_DEEP_ELF_SORCERER },
  {  0,  7,   45, SEMI, MONS_DEEP_ELF_DEATH_MAGE },
  {  6,  7,   10, UP,   MONS_DEEP_ELF_BLADEMASTER },
  {  6,  7,   10, UP,   MONS_DEEP_ELF_MASTER_ARCHER },
  { 0,0,0,FLAT,MONS_0 }
};

#if TAG_MAJOR_VERSION == 34
static const pop_entry pop_dwarf[] =
{ // Dwarven Hall
  {  1,  1, 1000, FLAT, MONS_DEEP_DWARF },
  {  1,  1,  690, FLAT, MONS_DEEP_DWARF_BERSERKER },
  {  1,  1,  690, FLAT, MONS_DEEP_DWARF_DEATH_KNIGHT },
  {  1,  1,  334, FLAT, MONS_UNBORN },
  {  1,  1,    3, FLAT, MONS_IRON_TROLL },
  {  1,  1,    3, FLAT, MONS_DEEP_TROLL },
  {  1,  1,    3, FLAT, MONS_DEEP_TROLL_EARTH_MAGE },
  {  1,  1,    3, FLAT, MONS_DEEP_TROLL_SHAMAN },
  {  1,  1,    8, FLAT, MONS_STONE_GIANT },
  {  1,  1,    8, FLAT, MONS_FIRE_GIANT },
  {  1,  1,    8, FLAT, MONS_FROST_GIANT },
  {  1,  1,  192, FLAT, MONS_WRAITH },
  {  1,  1,    3, FLAT, MONS_SHADOW_WRAITH },
  {  1,  1,    8, FLAT, MONS_EIDOLON },
  {  1,  1,    8, FLAT, MONS_PHANTASMAL_WARRIOR },
  { 0,0,0,FLAT,MONS_0 }
};
#endif

static const pop_entry pop_lair[] =
{ // Lair
  {  1,  5,   89, DOWN, MONS_GIANT_NEWT },
  { -4,  6,  514, SEMI, MONS_GIANT_GECKO },
  { -4,  6,  826, SEMI, MONS_IGUANA },
  {  1,  6,  999, FLAT, MONS_BAT },
  {  1,  6, 1000, FLAT, MONS_RAT },
  {  1,  6,  999, FLAT, MONS_QUOKKA },
  { -4,  6,  826, SEMI, MONS_JACKAL },
  { -4,  6,  514, SEMI, MONS_GIANT_CENTIPEDE },
  {  1,  7,  999, FLAT, MONS_PORCUPINE },
  { -3,  7,  690, SEMI, MONS_HOUND },
  {  1,  7,  999, FLAT, MONS_GIANT_FROG },
  { -1,  5,   52, SEMI, MONS_BLACK_BEAR },
  { -3,  7,  192, PEAK, MONS_WORM },
  {  1,  8,  923, FLAT, MONS_ADDER },
  { -2,  8,  749, SEMI, MONS_GREEN_RAT },
  {  1,  8,  981, FLAT, MONS_CROCODILE },
  { -2,  8,  192, PEAK, MONS_GIANT_MITE },
  {  0,  6,   25, PEAK, MONS_SCORPION },
  {  0,  8,  258, SEMI, MONS_BASILISK },
  {  1,  8,  981, FLAT, MONS_WATER_MOCCASIN },
  {  0,  8,  258, SEMI, MONS_WOLF },
  {  1,  8,  607, FLAT, MONS_GIANT_SLUG },
  {  1,  8,  514, FLAT, MONS_HIPPOGRIFF },
  {  1,  8,  879, FLAT, MONS_SPINY_FROG },
  {  2,  6,    8, PEAK, MONS_YELLOW_WASP },
  {  1,  7,   25, PEAK, MONS_GOLIATH_BEETLE },
  {  1,  8,  514, FLAT, MONS_KOMODO_DRAGON },
  {  2,  8,   25, PEAK, MONS_ORANGE_RAT },
  {  1,  8,  624, FLAT, MONS_AGATE_SNAIL },
  {  1,  9,  273, SEMI, MONS_SHEEP },
  {  1,  8,  514, FLAT, MONS_YAK },
  {  1,  9,  420, SEMI, MONS_BLINK_FROG },
  {  2,  8,   52, SEMI, MONS_GRIZZLY_BEAR },
  {  2,  8,   25, PEAK, MONS_STEAM_DRAGON },
  {  4,  6,    3, SEMI, MONS_GIANT_SPORE },
  {  1, 11,  457, SEMI, MONS_BLACK_MAMBA },
  {  1, 11,  514, SEMI, MONS_HYDRA },
  {  3,  9,   52, SEMI, MONS_CATOBLEPAS },
  {  4,  8,   14, SEMI, MONS_BRAIN_WORM },
  {  4,  8,    8, PEAK, MONS_ROCK_WORM },
  {  2, 10,   89, PEAK, MONS_WYVERN },
  {  1, 11,  273, PEAK, MONS_FIRE_DRAKE },
  {  4,  8,    8, PEAK, MONS_TRAPDOOR_SPIDER },
  {  3,  9,   25, PEAK, MONS_OKLOB_PLANT },
  {  2, 12,  192, PEAK, MONS_ELEPHANT_SLUG },
  {  2, 12,  273, PEAK, MONS_ELEPHANT },
  {  2, 12,  192, PEAK, MONS_GRIFFON },
  {  4, 10,   52, SEMI, MONS_POLAR_BEAR },
  {  4, 10,   25, PEAK, MONS_LINDWURM },
  {  5,  9,   11, PEAK, MONS_WOLF_SPIDER },
  {  5,  9,   17, SEMI, MONS_REDBACK },
  {  5,  9,   17, SEMI, MONS_WANDERING_MUSHROOM },
  {  3,  8,  192, UP  , MONS_DEATH_YAK },
  {  6, 10,   21, SEMI, MONS_SPINY_WORM },
  {  8,  8,    1, FLAT, MONS_FIRE_CRAB },
  {  3,  8,  180, UP  , MONS_BORING_BEETLE },
  {  4,  8,   89, UP  , MONS_BOULDER_BEETLE },
  { 0,0,0,FLAT,MONS_0 }
};

static const pop_entry pop_swamp[] =
{ // Swamp
  {  1,  5,  999, FLAT, MONS_FUNGUS },
  {  1,  5, 1000, FLAT, MONS_PLANT },
  {  1,  5,   25, DOWN, MONS_GIANT_NEWT },
  {  1,  5,  879, FLAT, MONS_WATER_MOCCASIN },
  {  1,  5,   25, DOWN, MONS_GIANT_SLUG },
  {  1,  5, 1000, FLAT, MONS_ALLIGATOR },
  {  1,  5,  192, FLAT, MONS_GIANT_FROG },
  {  1,  5,  923, FLAT, MONS_SWAMP_DRAKE },
  { -1,  5,   89, SEMI, MONS_GIANT_AMOEBA },
  { -1,  5,   89, SEMI, MONS_RAVEN },
  {  1,  5,   25, SEMI, MONS_AGATE_SNAIL },
  {  1,  5,  136, FLAT, MONS_CROCODILE },
  {  1,  5,  826, FLAT, MONS_HYDRA },
  {  1,  5,  999, FLAT, MONS_VAMPIRE_MOSQUITO },
  {  1,  5,  923, FLAT, MONS_BOG_BODY },
  {  1,  7,   52, SEMI, MONS_KOMODO_DRAGON },
  {  1,  5,  826, FLAT, MONS_GIANT_LEECH },
  {  1,  5, 1000, FLAT, MONS_SPINY_FROG },
  {  1,  5,  384, FLAT, MONS_SWAMP_DRAGON },
  {  1,  5,  192, FLAT, MONS_RED_WASP },
  {  1,  7,   40, PEAK, MONS_UGLY_THING },
  {  1,  5,  384, FLAT, MONS_INSUBSTANTIAL_WISP },
  {  1,  7,   40, PEAK, MONS_PHANTOM },
  {  1,  7,   40, PEAK, MONS_HUNGRY_GHOST },
  {  2,  8,   25, PEAK, MONS_BLINK_FROG },
  { -5,  5,  552, UP,   MONS_SLIME_CREATURE },
  {  2,  8,   52, SEMI, MONS_VERY_UGLY_THING },
  {  2,  8,   52, SEMI, MONS_VAPOUR },
  {  2,  5,   25, UP,   MONS_TENTACLED_MONSTROSITY },
  { 0,0,0,FLAT,MONS_0 }
};

static const pop_entry pop_shoals[] =
{ // Shoals
  {  1,  6,  258, DOWN, MONS_BAT },
  {  1,  5,  334, FLAT, MONS_SEA_SNAKE },
  {  1,  5,  258, FLAT, MONS_HIPPOGRIFF },
  {  1,  5,  981, FLAT, MONS_MERFOLK },
  {  1,  5,  334, FLAT, MONS_MERMAID },
  { -1,  5,   89, SEMI, MONS_CENTAUR },
  {  1,  5,  334, FLAT, MONS_MANTICORE },
  {  1,  5,  125, FLAT, MONS_HARPY },
  {  1,  5,  420, FLAT, MONS_SNAPPING_TURTLE },
  {  1,  7,   25, PEAK, MONS_OKLOB_PLANT },
  {  1,  7,  125, SEMI, MONS_SIREN },
  {  1,  7,   89, SEMI, MONS_CYCLOPS },
  {  1,  7,  258, FLAT, MONS_SHARK },
  {  1,  7,   73, SEMI, MONS_KRAKEN },
  {  2,  8,   25, PEAK, MONS_ALLIGATOR_SNAPPING_TURTLE },
  {  2,  8,   52, SEMI, MONS_MERFOLK_IMPALER },
  {  2,  8,   52, SEMI, MONS_MERFOLK_AQUAMANCER },
  {  2,  8,   25, PEAK, MONS_MERFOLK_JAVELINEER },
  { 0,0,0,FLAT,MONS_0 }
};

static const pop_entry pop_snake[] =
{ // Snake Pit
  { -1,  5,   52, SEMI, MONS_BALL_PYTHON },
  {  1,  5,  999, FLAT, MONS_ADDER },
  {  1,  5,  999, FLAT, MONS_WATER_MOCCASIN },
  {  1,  5,  849, FLAT, MONS_BLACK_MAMBA },
  {  1,  5,  217, FLAT, MONS_ANACONDA },
  {  1,  5,  571, FLAT, MONS_NAGA },
  {  1,  7,  244, FLAT, MONS_NAGA_WARRIOR },
  {  1,  7,  244, FLAT, MONS_NAGA_MAGE },
  {  2,  8,   52, SEMI, MONS_GUARDIAN_SERPENT },
  {  3,  9,   52, SEMI, MONS_GREATER_NAGA },
  { 0,0,0,FLAT,MONS_0 }
};

static const pop_entry pop_spider[] =
{ // Spider Nest
  { -1,  5,   89, SEMI, MONS_WORM },
  { -1,  5,   52, SEMI, MONS_GIANT_COCKROACH },
  { -1,  5,   52, SEMI, MONS_GIANT_CENTIPEDE },
  { -1,  5,   52, SEMI, MONS_GIANT_MITE },
  {  1,  5,  334, FLAT, MONS_SPIDER },
  {  1,  5,  258, FLAT, MONS_SCORPION },
  {  1,  5,   89, FLAT, MONS_ROCK_WORM },
  {  1,  5,   52, FLAT, MONS_YELLOW_WASP },
  {  1,  5,   89, FLAT, MONS_GOLIATH_BEETLE },
  {  1,  5,  879, FLAT, MONS_TRAPDOOR_SPIDER },
  {  1,  5,  607, FLAT, MONS_REDBACK },
  {  1,  7,   52, SEMI, MONS_SPINY_WORM },
  {  1,  7,   52, SEMI, MONS_BORING_BEETLE },
  {  1,  7,   89, SEMI, MONS_BOULDER_BEETLE },
  {  1,  7,  763, FLAT, MONS_JUMPING_SPIDER },
  {  1,  7,  552, FLAT, MONS_ORB_SPIDER },
  {  1,  7,  690, FLAT, MONS_TARANTELLA },
  {  0,  5,  258, UP,   MONS_RED_WASP },
  {  1,  5,  957, FLAT, MONS_WOLF_SPIDER },
  {  1,  5,   89, UP,   MONS_EMPEROR_SCORPION },
  {  0, 10,  763, SEMI, MONS_DEMONIC_CRAWLER },
  {  2, 10,   89, PEAK, MONS_GHOST_MOTH },
  {  2, 10,   89, PEAK, MONS_MOTH_OF_SUPPRESSION },
  {  3,  9,   52, SEMI, MONS_MOTH_OF_WRATH },
  { 0,0,0,FLAT,MONS_0 }
};

static const pop_entry pop_slime[] =
{ // Slime Pits
  {  1,  6,  514, FLAT, MONS_OOZE },
  {  1,  6, 1000, FLAT, MONS_JELLY },
  {  1,  6, 1000, FLAT, MONS_ACID_BLOB },
  {  1,  6,  514, FLAT, MONS_GIANT_EYEBALL },
  {  0,  6, 1000, FLAT, MONS_BROWN_OOZE },
  {  0,  6, 1000, FLAT, MONS_SLIME_CREATURE },
  {  0,  6,  514, FLAT, MONS_EYE_OF_DRAINING },
  {  1,  6, 1000, FLAT, MONS_GIANT_AMOEBA },
  {  1,  6,  514, FLAT, MONS_AZURE_JELLY },
  {  1,  6,  514, FLAT, MONS_SHINING_EYE },
  {  1,  6,  192, FLAT, MONS_GOLDEN_EYE },
  {  1,  9,   89, PEAK, MONS_PULSATING_LUMP },
  {  1,  9,  192, SEMI, MONS_EYE_OF_DEVASTATION },
  {  1,  9,  192, SEMI, MONS_GREAT_ORB_OF_EYES },
  {  1,  6,  192, UP,   MONS_DEATH_OOZE },
  {  2,  6,   89, UP,   MONS_GIANT_ORANGE_BRAIN },
  { 0,0,0,FLAT,MONS_0 }
};

static const pop_entry pop_vaults[] =
{ // Vaults (OOD cap: 12)
  {  1,  1,   32, FLAT, MONS_KOMODO_DRAGON },
  {  1,  1,   43, FLAT, MONS_SOLDIER_ANT },
  {  1,  1,   38, FLAT, MONS_TRAPDOOR_SPIDER },
  {  1,  1,  999, FLAT, MONS_HILL_GIANT },
  {  1,  1,  257, FLAT, MONS_UNSEEN_HORROR },
  {  1,  1,  187, FLAT, MONS_WRAITH },
  {  1,  2,   86, DOWN, MONS_BASILISK },
  {  1,  2,  282, FLAT, MONS_HYDRA },
  {  1,  2,   49, DOWN, MONS_BRAIN_WORM },
  { -8,  2,  334, SEMI, MONS_MOTTLED_DRAGON },
  { -8,  2,  826, SEMI, MONS_CYCLOPS },
  {  1,  2,   64, DOWN, MONS_SKELETAL_WARRIOR },
  {  1,  1,   26, FLAT, MONS_BLINK_FROG },
  { -7,  3,  879, SEMI, MONS_SLIME_CREATURE },
  { -7,  3,  514, SEMI, MONS_VAMPIRE_MOSQUITO },
  {  1,  1,    8, FLAT, MONS_STONE_GOLEM },
  { -5,  1,   25, PEAK, MONS_WANDERING_MUSHROOM },
  {  1,  2,   36, DOWN, MONS_RAKSHASA },
  { -2,  3,  192, DOWN, MONS_VAMPIRE },
  { -2,  3,  999, FLAT, MONS_ZOMBIE },
  { -6,  4,  334, SEMI, MONS_GRIFFON },
  { -6,  4,  230, PEAK, MONS_ROCK_WORM },
  {  1,  4,  999, FLAT, MONS_UGLY_THING },
  { -4,  2,   52, SEMI, MONS_CLAY_GOLEM },
  { -5,  3,  136, SEMI, MONS_BOGGART },
  {  1,  4,  923, FLAT, MONS_CENTAUR_WARRIOR },
  { -5,  3,   89, PEAK, MONS_WIZARD },
  { -6,  4,  674, SEMI, MONS_SHAPESHIFTER },
  { -5,  3,  272, SEMI, MONS_SIMULACRUM },
  { -5,  5,  192, PEAK, MONS_DRAGON },
  { -5,  5,  258, PEAK, MONS_GIANT_AMOEBA },
  { -2,  2,   11, PEAK, MONS_GARGOYLE },
  { -3,  3,   40, PEAK, MONS_KOBOLD_DEMONOLOGIST },
  {  1,  5,  999, FLAT, MONS_SKELETON },
  {  1,  6,  217, DOWN, MONS_GIANT_SLUG },
  {  1,  6,  273, DOWN, MONS_WOLF_SPIDER },
  {  1,  4,   25, DOWN, MONS_IRON_GOLEM },
  {  0,  2,    3, SEMI, MONS_TOENAIL_GOLEM },
  { -4,  6,  334, SEMI, MONS_YAKTAUR },
  { -4,  6,  420, SEMI, MONS_OGRE_MAGE },
  { -2,  6,   89, PEAK, MONS_AGATE_SNAIL },
  { -3,  7,  192, PEAK, MONS_SPINY_WORM },
  {  0, 10,  367, FLAT, MONS_VAULT_WARDEN },
  { -2,  6,   89, PEAK, MONS_ICE_DRAGON },
  { -2,  6,  136, SEMI, MONS_GREAT_ORB_OF_EYES },
  { -1,  5,   25, PEAK, MONS_GIANT_ORANGE_BRAIN },
  { -2,  6,   89, PEAK, MONS_HARPY },
  { -2,  6,   89, PEAK, MONS_FIRE_CRAB },
  { -2,  6,  136, SEMI, MONS_RED_WASP },
  { -2,  6,   89, PEAK, MONS_VERY_UGLY_THING },
  {  1,  3,    6, FLAT, MONS_EARTH_ELEMENTAL },
  {  1,  3,    6, FLAT, MONS_FIRE_ELEMENTAL },
  {  1,  3,    6, FLAT, MONS_AIR_ELEMENTAL },
  { -3,  7,  690, SEMI, MONS_ORC_KNIGHT },
  { -2,  6,  136, SEMI, MONS_DEEP_ELF_FIGHTER },
  { -2,  6,  136, SEMI, MONS_DEEP_ELF_KNIGHT },
  {  0,  4,    8, PEAK, MONS_DEEP_ELF_MAGE },
  {  0,  4,    8, PEAK, MONS_DEEP_ELF_SUMMONER },
  {  0,  4,    8, PEAK, MONS_NAGA_WARRIOR },
  {  0,  4,    8, PEAK, MONS_NAGA_MAGE },
  { -1,  5,   52, SEMI, MONS_NECROMANCER },
  { -1,  5,   25, PEAK, MONS_SHADOW_WRAITH },
  {  2,  4,    3, SEMI, MONS_PULSATING_LUMP },
  { -2,  8,  244, PEAK, MONS_BOULDER_BEETLE },
  {  0,  6,   35, PEAK, MONS_ORC_HIGH_PRIEST },
  {  0,  8,   89, PEAK, MONS_REDBACK },
  {  0,  8,   89, PEAK, MONS_SPHINX },
  {  2,  6,    8, PEAK, MONS_CRYSTAL_GOLEM },
  {  2,  6,    8, PEAK, MONS_VAPOUR },
  { -1,  9,   60, PEAK, MONS_FLAYED_GHOST },
  { -1,  9,  258, PEAK, MONS_FREEZING_WRAITH },
  { -3, 10,  480, SEMI, MONS_VAULT_SENTINEL },
  { -3, 10,  480, SEMI, MONS_IRONBRAND_CONVOKER },
  {  4,  6,    3, SEMI, MONS_SHINING_EYE },
  {  2,  8,   25, PEAK, MONS_ORC_SORCERER },
  {  4,  8,   14, SEMI, MONS_ORC_WARLORD },
  {  1, 11,  826, SEMI, MONS_YAKTAUR_CAPTAIN },
  {  2, 10,  136, SEMI, MONS_IRON_TROLL },
  {  5,  9,    8, PEAK, MONS_DANCING_WEAPON },
  {  3, 11,  136, SEMI, MONS_DEEP_TROLL },
  {  2, 12,  192, PEAK, MONS_ETTIN },
  {  2, 12,  571, SEMI, MONS_STONE_GIANT },
  {  0, 11,  420, SEMI, MONS_IRONHEART_PRESERVER },
  {  3, 11,  136, SEMI, MONS_FIRE_GIANT },
  {  3, 11,  136, SEMI, MONS_FROST_GIANT },
  {  4, 12,  155, SEMI, MONS_TENGU_REAVER },
  {  4, 10,   25, PEAK, MONS_HELL_KNIGHT },
  {  3, 11,   66, PEAK, MONS_LICH },
  {  4, 12,   89, PEAK, MONS_STORM_DRAGON },
  {  7,  9,    6, FLAT, MONS_DEEP_ELF_CONJURER },
  {  4, 12,   89, PEAK, MONS_PHANTASMAL_WARRIOR },
  {  8, 10,    4, SEMI, MONS_DEEP_ELF_PRIEST },
  {  4, 14,  258, PEAK, MONS_GLOWING_SHAPESHIFTER },
  {  5, 13,   66, PEAK, MONS_TENTACLED_MONSTROSITY },
  {  8, 12,   89, UP,   MONS_SHADOW_DRAGON },
  { 10, 12,    8, UP,   MONS_BONE_DRAGON },
  { 10, 12,    8, UP,   MONS_QUICKSILVER_DRAGON },
  { 10, 12,    8, UP,   MONS_IRON_DRAGON },
  { 10, 14,   14, SEMI, MONS_GOLDEN_DRAGON },
  { 10, 12,    8, UP,   MONS_DEEP_ELF_HIGH_PRIEST },
  { 11, 12,    4, UP,   MONS_DEEP_ELF_DEMONOLOGIST },
  { 11, 12,    4, UP,   MONS_DEEP_ELF_ANNIHILATOR },
  { 11, 12,    4, UP,   MONS_DEEP_ELF_SORCERER },
  { 11, 12,    4, UP,   MONS_DEEP_ELF_DEATH_MAGE },
  {  9, 15,   25, PEAK, MONS_TITAN },
  { 11, 12,    4, UP,   MONS_ANCIENT_LICH },
  { 0,0,0,FLAT,MONS_0 }
};

static const pop_entry pop_blade[] =
{ // Hall of Blades
  {  1,  1, 1000, FLAT, MONS_DANCING_WEAPON },
  { 0,0,0,FLAT,MONS_0 }
};

static const pop_entry pop_crypt[] =
{ // Crypt
  {  2,  7,  145, SEMI, MONS_EIDOLON },
  {  1,  6,  160, DOWN, MONS_WRAITH },
  {  1,  3,   65, SEMI, MONS_FREEZING_WRAITH },
  {  1,  5,  125, SEMI, MONS_PHANTASMAL_WARRIOR },
  {  3,  6,   95, SEMI, MONS_SHADOW_WRAITH },
  {  1,  5,   75, FLAT, MONS_SKELETAL_WARRIOR },
  {  0,  6,  140, UP,   MONS_ANCIENT_CHAMPION },
  {  1,  4,   25, FLAT, MONS_FLAMING_CORPSE },
  {  2,  5,   40, SEMI, MONS_GHOUL },
  {  1,  5,   30, FLAT, MONS_PLAGUE_SHAMBLER },
  {  1,  4,   85, DOWN, MONS_VAMPIRE },
  {  1,  6,  125, PEAK, MONS_VAMPIRE_MAGE },
  {  1,  6,  105, UP,   MONS_VAMPIRE_KNIGHT },
  {  1,  5,   75, SEMI, MONS_JIANGSHI },
  { -2,  4,   75, SEMI, MONS_NECROMANCER },
  {  1,  5,   30, FLAT, MONS_SILENT_SPECTRE },
  {  3,  6,  145, SEMI, MONS_FLAYED_GHOST },
  {  3,  8,   40, SEMI, MONS_REAPER },
  {  2,  6,   55, SEMI, MONS_SOUL_EATER },
  {  2,  9,   70, UP,   MONS_BONE_DRAGON },
  {  1,  7,   55, SEMI, MONS_DEEP_DWARF_DEATH_KNIGHT },
  {  2,  7,   80, SEMI, MONS_DEEP_ELF_DEATH_MAGE },
  {  3,  7,   40, SEMI, MONS_UNBORN },
  {  2,  8,   85, SEMI, MONS_REVENANT },
  {  3,  8,   55, PEAK, MONS_LICH },
  {  4,  8,   15, PEAK, MONS_ANCIENT_LICH },
  { -2,  4,   70, PEAK, MONS_SHADOW },
  { -2,  4,   40, PEAK, MONS_WIGHT },
  { -2,  4,   45, PEAK, MONS_FLYING_SKULL },
  {  1,  5,   35, FLAT, MONS_HUNGRY_GHOST },
  { 0,0,0,FLAT,MONS_0 }
};

static const pop_entry pop_tomb[] =
{ // Tomb (OOD cap: 5)
  {  1,  5,  187, FLAT, MONS_ZOMBIE },
  {  1,  3,    3, SEMI, MONS_TRAPDOOR_SPIDER },
  {  1,  5, 1000, FLAT, MONS_MUMMY },
  {  1,  5,  199, FLAT, MONS_SKELETON },
  {  1,  5, 1000, FLAT, MONS_GUARDIAN_MUMMY },
  {  1,  5,  230, FLAT, MONS_FLYING_SKULL },
  {  1,  5,   50, SEMI, MONS_SIMULACRUM },
  { -1,  9,  334, SEMI, MONS_MUMMY_PRIEST },
  {  3,  5,    6, FLAT, MONS_LICH },
  {  3,  5,    3, SEMI, MONS_ANCIENT_LICH },
  {  4,  5,    3, UP,   MONS_GREATER_MUMMY },
  { 0,0,0,FLAT,MONS_0 }
};

static const pop_entry pop_hell[] =
{ // Hell
  {  1,  1,   89, FLAT, MONS_ORANGE_RAT },
  {  1,  1,   89, FLAT, MONS_HELL_HOUND },
  {  1,  1,  136, FLAT, MONS_HELL_HOG },
  {  1,  1,   25, FLAT, MONS_HELLEPHANT },
  {  1,  1,   89, FLAT, MONS_DEMONIC_CRAWLER },
  {  1,  1,  136, FLAT, MONS_HELL_KNIGHT },
  {  1,  1,  136, FLAT, MONS_NECROMANCER },
  {  1,  1,   89, FLAT, MONS_RED_DEVIL },
  {  1,  1,   25, FLAT, MONS_HELLION },
  {  1,  1,   89, FLAT, MONS_LOROCYPROCA },
  {  1,  1,   25, FLAT, MONS_TORMENTOR },
  {  1,  1,  514, FLAT, MONS_REAPER },
  {  1,  1,  514, FLAT, MONS_SOUL_EATER },
  {  1,  1,  514, FLAT, MONS_ICE_DEVIL },
  {  1,  1,   89, FLAT, MONS_BLUE_DEVIL },
  {  1,  1,   89, FLAT, MONS_IRON_DEVIL },
  {  1,  1,   52, FLAT, MONS_GREEN_DEATH },
  {  1,  1,   52, FLAT, MONS_BLIZZARD_DEMON },
  {  1,  1,   52, FLAT, MONS_BALRUG },
  {  1,  1,   52, FLAT, MONS_CACODEMON },
  {  1,  1,  514, FLAT, MONS_SUN_DEMON },
  {  1,  1,   25, FLAT, MONS_TENTACLED_MONSTROSITY },
  { 0,0,0,FLAT,MONS_0 }
};

static const pop_entry pop_dis[] =
{ // Dis
  {  1,  7,   25, DOWN, MONS_CLAY_GOLEM },
  {  1,  2,    8, DOWN, MONS_CRIMSON_IMP },
  {  1,  7,  514, DOWN, MONS_RED_DEVIL },
  {  1,  7,  136, DOWN, MONS_NECROPHAGE },
  {  1, 12,  514, DOWN, MONS_SKELETAL_WARRIOR },
  {  1,  7, 1998, FLAT, MONS_ZOMBIE },
  {  1, 12,  438, DOWN, MONS_HELL_HOUND },
  {  1,  7,   25, DOWN, MONS_STONE_GOLEM },
  {  1,  7,   25, DOWN, MONS_HELL_KNIGHT },
  {  1,  7,   25, DOWN, MONS_ROTTING_DEVIL },
  {  1,  7,  776, FLAT, MONS_TORMENTOR },
  {  1,  7,   66, DOWN, MONS_PHANTOM },
  {  1,  5,  624, DOWN, MONS_SHADOW },
  {  1,  6,  420, DOWN, MONS_WIGHT },
  {  1, 12,  272, DOWN, MONS_SKELETON },
  {  1,  7,   89, DOWN, MONS_HELL_HOG },
  {  1, 12,   25, DOWN, MONS_IRON_GOLEM },
  {  1,  7,  300, FLAT, MONS_HELLION },
  {  1,  7,   25, FLAT, MONS_EFREET },
  {  1,  7,  420, DOWN, MONS_MUMMY },
  {  1,  7,  258, DOWN, MONS_FLYING_SKULL },
  {  1,  7,  999, FLAT, MONS_IRON_IMP },
  {  1,  7,  999, FLAT, MONS_IRON_DEVIL },
  {  1,  7,   81, PEAK, MONS_VAMPIRE },
  {  1,  7,  136, SEMI, MONS_FLAYED_GHOST },
  {  1,  7,  476, SEMI, MONS_WRAITH },
  {  1,  7,  192, SEMI, MONS_FREEZING_WRAITH },
  {  1,  7,   25, PEAK, MONS_DANCING_WEAPON },
  {  1,  7,  898, FLAT, MONS_REAPER },
  {  1,  7,   25, PEAK, MONS_SOUL_EATER },
  {  1,  7,  192, FLAT, MONS_ICE_DEVIL },
  {  1,  7,  334, FLAT, MONS_BLUE_DEVIL },
  {  1,  7,  136, SEMI, MONS_FLAMING_CORPSE },
  {  1,  7,   25, PEAK, MONS_PHANTASMAL_WARRIOR },
  {  1,  7,   25, PEAK, MONS_LICH },
  {  1,  9,   89, PEAK, MONS_BONE_DRAGON },
  {  2,  8,   25, PEAK, MONS_IRON_DRAGON },
  {  4,  7,    4, FLAT, MONS_BRIMSTONE_FIEND },
//{  4,  6,    4, SEMI, MONS_ANCIENT_LICH },
  { 0,0,0,FLAT,MONS_0 }
};

static const pop_entry pop_geh[] =
{ // Gehenna
  {  1,  7,   25, DOWN, MONS_CLAY_GOLEM },
  {  1, 12,  690, DOWN, MONS_RED_DEVIL },
  {  1,  7, 1998, FLAT, MONS_SKELETON },
  {  1,  7, 1998, FLAT, MONS_ZOMBIE },
  {  1,  7,  350, FLAT, MONS_HELL_HOUND },
  {  1,  7,  192, FLAT, MONS_HELL_HOG },
  {  1,  7,   17, FLAT, MONS_STONE_GOLEM },
  {  1,  7,  192, DOWN, MONS_CRIMSON_IMP },
  {  1,  7,  514, DOWN, MONS_NECROPHAGE },
  {  0,  4,    8, PEAK, MONS_IRON_GOLEM },
  { -3,  7,  514, SEMI, MONS_ROTTING_DEVIL },
  {  1,  7,  217, DOWN, MONS_PHANTOM },
  {  1,  7,  705, DOWN, MONS_SHADOW },
  {  1,  7,  690, DOWN, MONS_WIGHT },
  {  1,  7,  384, FLAT, MONS_FLYING_SKULL },
  { -1,  7,   98, PEAK, MONS_HELL_KNIGHT },
  { -1,  7,   89, PEAK, MONS_IRON_DEVIL },
  { -1,  7,  402, SEMI, MONS_VAMPIRE },
  { -1,  7,  571, FLAT, MONS_WRAITH },
  {  2,  6,    8, PEAK, MONS_FIRE_CRAB },
  {  1,  7,  589, FLAT, MONS_HELLION },
  {  1,  7,  367, FLAT, MONS_TORMENTOR },
  {  1,  7,  192, FLAT, MONS_BALRUG },
  {  2,  6,    8, PEAK, MONS_EFREET },
  {  1,  7,  217, FLAT, MONS_FLAMING_CORPSE },
  {  1,  7,  217, FLAT, MONS_FLAYED_GHOST },
  {  2,  8,   25, PEAK, MONS_BONE_DRAGON },
  {  1,  9,  384, SEMI, MONS_REAPER },
  {  3,  7,    8, PEAK, MONS_SOUL_EATER },
  {  3,  7,   14, SEMI, MONS_HELL_SENTINEL },
  {  3,  7,    8, PEAK, MONS_BRIMSTONE_FIEND },
  {  1,  9,   89, PEAK, MONS_PHANTASMAL_WARRIOR },
  {  1,  9,  136, SEMI, MONS_LICH },
  {  4,  7,    2, FLAT, MONS_ANCIENT_LICH },
  { 0,0,0,FLAT,MONS_0 }
};

static const pop_entry pop_coc[] =
{ // Cocytus
  {  1,  7,  136, DOWN, MONS_NECROPHAGE },
  {  1,  7,  514, FLAT, MONS_SKELETAL_WARRIOR },
  {  1,  7, 1914, FLAT, MONS_SKELETON },
  {  1,  7, 1914, FLAT, MONS_ZOMBIE },
  {  1,  7,   70, DOWN, MONS_SIMULACRUM },
  {  1,  7,  957, DOWN, MONS_ICE_BEAST },
  {  1,  7,  889, SEMI, MONS_BLUE_DEVIL },
  {  1,  5,  136, DOWN, MONS_PHANTOM },
  {  1,  6,  624, DOWN, MONS_SHADOW },
  {  1,  7,  287, DOWN, MONS_ROTTING_DEVIL },
  {  1,  7,  244, DOWN, MONS_VAMPIRE },
  {  1,  7,  420, DOWN, MONS_WIGHT },
  {  1,  7,  641, SEMI, MONS_FLYING_SKULL },
  { -1,  7,  258, SEMI, MONS_MUMMY },
  { -1,  7,  146, SEMI, MONS_HUNGRY_GHOST },
  { -1,  7,  420, SEMI, MONS_WRAITH },
  {  1,  7,  968, FLAT, MONS_FREEZING_WRAITH },
  {  1,  7,   89, PEAK, MONS_PHANTASMAL_WARRIOR },
  {  1,  7,  303, FLAT, MONS_ICE_DRAGON },
  {  1,  7,  889, FLAT, MONS_WHITE_IMP },
  {  1,  7,  287, FLAT, MONS_TORMENTOR },
  {  1,  7,  889, FLAT, MONS_ICE_DEVIL },
  {  1,  7,  192, FLAT, MONS_BLIZZARD_DEMON },
  {  2,  7,   35, FLAT, MONS_BONE_DRAGON },
  {  1,  7,  457, UP,   MONS_REAPER },
  {  1,  7,   81, PEAK, MONS_SOUL_EATER },
  {  2,  8,   35, PEAK, MONS_LICH },
  {  3,  7,    8, PEAK, MONS_ANCIENT_LICH },
  { 0,0,0,FLAT,MONS_0 }
};

static const pop_entry pop_tar[] =
{ // Tartarus
  {  1,  7,   89, DOWN, MONS_CRIMSON_IMP },
  {  1,  7,  999, FLAT, MONS_SHADOW_IMP },
  {  1,  7,   40, PEAK, MONS_RED_DEVIL },
  {  1,  7,  999, FLAT, MONS_SKELETAL_WARRIOR },
  {  1,  7, 1998, FLAT, MONS_SKELETON },
  {  1,  7,   89, FLAT, MONS_DEATH_DRAKE },
  {  1,  7,   46, DOWN, MONS_HELL_KNIGHT },
  {  1,  7,  849, SEMI, MONS_NECROPHAGE },
  {  1,  7,  552, DOWN, MONS_PHANTOM },
  {  1,  7,  837, DOWN, MONS_WIGHT },
  {  1,  7, 1998, FLAT, MONS_ZOMBIE },
  {  1,  7,  552, DOWN, MONS_HELL_HOUND },
  {  1,  7,   35, PEAK, MONS_NECROMANCER },
  {  1,  7,  988, FLAT, MONS_SHADOW },
  {  1,  7,  552, SEMI, MONS_WRAITH },
  {  1,  7,  690, SEMI, MONS_FREEZING_WRAITH },
  {  1,  7,  258, FLAT, MONS_SILENT_SPECTRE },
  {  1,  7,  334, FLAT, MONS_BONE_DRAGON },
  {  1,  7,  367, FLAT, MONS_TORMENTOR },
  {  1,  7,  244, SEMI, MONS_ICE_DEVIL },
  {  1,  7,  217, SEMI, MONS_BLUE_DEVIL },
  {  1,  7,  230, SEMI, MONS_MUMMY },
  {  1,  7,  217, FLAT, MONS_HUNGRY_GHOST },
  {  1,  7,  192, FLAT, MONS_FLAYED_GHOST },
  {  1,  7,  420, FLAT, MONS_PHANTASMAL_WARRIOR },
  {  1,  7,   70, PEAK, MONS_SIMULACRUM },
  {  1,  7,   89, SEMI, MONS_SHADOW_DRAGON },
  {  1,  7,  720, FLAT, MONS_ROTTING_DEVIL },
  {  1,  7,  367, FLAT, MONS_HELLION },
  {  1,  7,  859, FLAT, MONS_REAPER },
  {  1,  7,  367, FLAT, MONS_SHADOW_DEMON },
  {  1,  7,  402, FLAT, MONS_VAMPIRE },
  {  1,  7,  571, FLAT, MONS_FLYING_SKULL },
  {  1,  9,  258, SEMI, MONS_SOUL_EATER },
  {  1,  9,  125, PEAK, MONS_LICH },
  {  3,  7,   11, PEAK, MONS_ANCIENT_LICH },
  { 0,0,0,FLAT,MONS_0 }
};

static const pop_entry pop_zot[] =
{ // Zot
  {  1,  5,   89, DOWN, MONS_ELECTRIC_GOLEM },
  {  1,  5,   42, FLAT, MONS_ORB_OF_FIRE },
  {  1,  5,  972, FLAT, MONS_MOTH_OF_WRATH },
  {  1,  5,   73, FLAT, MONS_BLACK_DRACONIAN },
  {  1,  5,   73, FLAT, MONS_MOTTLED_DRACONIAN },
  {  1,  5,   73, FLAT, MONS_YELLOW_DRACONIAN },
  {  1,  5,   73, FLAT, MONS_GREEN_DRACONIAN },
  {  1,  5,   73, FLAT, MONS_PURPLE_DRACONIAN },
  {  1,  5,   73, FLAT, MONS_RED_DRACONIAN },
  {  1,  5,   73, FLAT, MONS_WHITE_DRACONIAN },
  {  1,  5,   73, FLAT, MONS_GREY_DRACONIAN },
  {  1,  5,   73, FLAT, MONS_PALE_DRACONIAN },
  {  1,  5,   59, FLAT, MONS_DRACONIAN_CALLER },
  {  1,  5,   59, FLAT, MONS_DRACONIAN_MONK },
  {  1,  5,   59, FLAT, MONS_DRACONIAN_ZEALOT },
  {  1,  5,   59, FLAT, MONS_DRACONIAN_SHIFTER },
  {  1,  5,   59, FLAT, MONS_DRACONIAN_ANNIHILATOR },
  {  1,  5,   59, FLAT, MONS_DRACONIAN_KNIGHT },
  {  1,  5,   59, FLAT, MONS_DRACONIAN_SCORCHER },
  {  1,  5,  514, FLAT, MONS_TENTACLED_MONSTROSITY },
  {  1,  5,  334, FLAT, MONS_DRAGON },
  {  1,  5,  334, FLAT, MONS_ICE_DRAGON },
  {  1,  5,  334, FLAT, MONS_DEATH_COB },
  {  1,  5,  192, UP,   MONS_SHADOW_DRAGON },
  { -4,  5,  514, UP,   MONS_STORM_DRAGON },
  { -4,  5,  334, UP,   MONS_BONE_DRAGON },
  {  1,  5,  192, UP,   MONS_GHOST_MOTH },
  {  2,  8,   52, SEMI, MONS_KILLER_KLOWN },
  {  1,  5,  150, UP,   MONS_CURSE_TOE },
  {  1, 11,  367, SEMI, MONS_GOLDEN_DRAGON },
  { 0,0,0,FLAT,MONS_0 }
};

static const pop_entry pop_forest[] =
{ // Forest
  {  1,  5,  120, DOWN, MONS_WOLF },
  {  1,  6,   75, DOWN, MONS_GRIZZLY_BEAR },
  {  1,  5,   35, DOWN, MONS_BLACK_BEAR },
  {  1,  5,   50, FLAT, MONS_YAK },
  {  1,  7,  145, SEMI, MONS_DIRE_ELEPHANT },
  {  1,  5,   15, DOWN, MONS_YELLOW_WASP },
  {  1,  5,   30, FLAT, MONS_RED_WASP },
  {  1,  9,  100, SEMI, MONS_FIREFLY },
  {  1,  6,   75, DOWN, MONS_REDBACK },
  {  2,  7,   35, SEMI, MONS_WOLF_SPIDER },
  {  1,  9,   75, SEMI, MONS_OKLOB_PLANT },
  {  1,  5,  170, FLAT, MONS_DRYAD },
  {  1,  5,  120, FLAT, MONS_WIND_DRAKE },
  { -1,  5,   75, SEMI, MONS_FAUN },
  {  0,  9,  105, SEMI, MONS_SATYR },
  {  2,  8,   55, SEMI, MONS_SPRIGGAN_DRUID },
  {  1,  6,  155, SEMI, MONS_SPRIGGAN_RIDER },
  {  2,  6,  135, SEMI, MONS_SPRIGGAN_ASSASSIN },
  {  2,  8,  115, SEMI, MONS_SPRIGGAN_ENCHANTER },
  {  1,  9,  235, SEMI, MONS_SPRIGGAN_BERSERKER },
  {  1,  8,  155, SEMI, MONS_SPRIGGAN_AIR_MAGE },
  {  3,  5,  115, UP,   MONS_SPRIGGAN_DEFENDER },
  {  1,  7,   85, PEAK, MONS_APIS },
  {  2,  8,  200, SEMI, MONS_ANCIENT_BEAR },
  {  2,  7,  165, SEMI, MONS_TREANT },
  {  2,  5,  110, FLAT, MONS_SPIRIT_WOLF },
  {  1,  6,   80, SEMI, MONS_IRON_TROLL },
  {  1,  6,   85, SEMI, MONS_ANACONDA },
  {  1,  9,  100, PEAK, MONS_THORN_HUNTER },
  {  1,  5,  125, FLAT, MONS_BUTTERFLY },
  { 0,0,0,FLAT,MONS_0 }
};

static const pop_entry pop_abyss[] =
{ // Abyss
  {  1,  5,   46, FLAT, MONS_FIRE_BAT },
  {  1,  5,   30, FLAT, MONS_HELL_HOUND },
  {  1,  5,   30, FLAT, MONS_HELL_HOG },
  {  1,  5,    4, FLAT, MONS_HELLEPHANT },
  {  1,  5,   59, FLAT, MONS_BRAIN_WORM },
  {  1,  5,    8, FLAT, MONS_SPINY_WORM },
  {  1,  5,    8, FLAT, MONS_DEATH_DRAKE },
  {  1,  5,   89, FLAT, MONS_BONE_DRAGON },
  {  1,  5,   21, FLAT, MONS_JELLY },
  {  1,  5,   17, FLAT, MONS_GIANT_EYEBALL },
  {  1,  5,   46, FLAT, MONS_EYE_OF_DRAINING },
  {  1,  5,    8, FLAT, MONS_SHINING_EYE },
  {  1,  5,    4, FLAT, MONS_EYE_OF_DEVASTATION },
  {  1,  5,    8, FLAT, MONS_GREAT_ORB_OF_EYES },
  {  1,  5,   52, FLAT, MONS_GIANT_ORANGE_BRAIN },
  {  1,  5,    3, FLAT, MONS_DANCING_WEAPON },
  {  1,  5,   25, FLAT, MONS_DEMONIC_CRAWLER },
  {  1,  5,    4, FLAT, MONS_UGLY_THING },
  {  1,  5,   52, FLAT, MONS_VERY_UGLY_THING },
  {  1,  5,   21, FLAT, MONS_ICE_BEAST },
  {  1,  5,   21, FLAT, MONS_SKY_BEAST },
  {  1,  5,    8, FLAT, MONS_CLAY_GOLEM },
  {  1,  5,    8, FLAT, MONS_STONE_GOLEM },
  {  1,  5,    8, FLAT, MONS_IRON_GOLEM },
  {  1,  5,    4, FLAT, MONS_CRYSTAL_GOLEM },
  {  1,  5,    1, FLAT, MONS_TOENAIL_GOLEM },
  {  1,  5,    6, FLAT, MONS_EARTH_ELEMENTAL },
  {  1,  5,    6, FLAT, MONS_FIRE_ELEMENTAL },
  {  1,  5,    6, FLAT, MONS_AIR_ELEMENTAL },
  {  1,  5,   35, FLAT, MONS_GOLDEN_EYE },
  {  1,  5,   35, FLAT, MONS_INSUBSTANTIAL_WISP },
  {  1,  5,   21, FLAT, MONS_VAPOUR },
  {  1,  5,   17, FLAT, MONS_FUNGUS },
  {  1,  5,    3, FLAT, MONS_GIANT_SPORE },
  {  1,  5,    8, FLAT, MONS_DEEP_DWARF },
  {  1,  5,    8, FLAT, MONS_DEEP_DWARF_BERSERKER },
  {  1,  5,    8, FLAT, MONS_DEEP_DWARF_DEATH_KNIGHT },
  {  1,  5,    8, FLAT, MONS_UNBORN },
  {  1,  5,    8, FLAT, MONS_TENGU },
  {  1,  5,    8, FLAT, MONS_NAGA_WARRIOR },
  {  1,  5,    8, FLAT, MONS_NAGA_MAGE },
  {  1,  5,   25, FLAT, MONS_GUARDIAN_SERPENT },
  {  1,  5,    8, FLAT, MONS_OCTOPODE },
  {  1,  5,    4, FLAT, MONS_HELL_KNIGHT },
  {  1,  5,    4, FLAT, MONS_NECROMANCER },
  {  1,  5,    8, FLAT, MONS_WIZARD },
  {  1,  5,    4, FLAT, MONS_BLACK_DRACONIAN },
  {  1,  5,    4, FLAT, MONS_MOTTLED_DRACONIAN },
  {  1,  5,    4, FLAT, MONS_YELLOW_DRACONIAN },
  {  1,  5,    4, FLAT, MONS_GREEN_DRACONIAN },
  {  1,  5,    4, FLAT, MONS_PURPLE_DRACONIAN },
  {  1,  5,    4, FLAT, MONS_RED_DRACONIAN },
  {  1,  5,    4, FLAT, MONS_WHITE_DRACONIAN },
  {  1,  5,    4, FLAT, MONS_GREY_DRACONIAN },
  {  1,  5,    4, FLAT, MONS_PALE_DRACONIAN },
  {  1,  5,    4, FLAT, MONS_DRACONIAN_CALLER },
  {  1,  5,    4, FLAT, MONS_DRACONIAN_MONK },
  {  1,  5,    4, FLAT, MONS_DRACONIAN_ZEALOT },
  {  1,  5,    4, FLAT, MONS_DRACONIAN_SHIFTER },
  {  1,  5,    4, FLAT, MONS_DRACONIAN_ANNIHILATOR },
  {  1,  5,    4, FLAT, MONS_DRACONIAN_KNIGHT },
  {  1,  5,    4, FLAT, MONS_DRACONIAN_SCORCHER },
  {  1,  5,   25, FLAT, MONS_CRIMSON_IMP },
  {  1,  5,  192, FLAT, MONS_WHITE_IMP },
  {  1,  5,  192, FLAT, MONS_LEMURE },
  {  1,  5,  192, FLAT, MONS_UFETUBUS },
  {  1,  5,  192, FLAT, MONS_IRON_IMP },
  {  1,  5,   25, FLAT, MONS_SHADOW_IMP },
  {  1,  5,    8, FLAT, MONS_RED_DEVIL },
  {  1,  5,   25, FLAT, MONS_ROTTING_DEVIL },
  {  1,  5,  514, FLAT, MONS_HELLWING },
  {  1,  5,   89, FLAT, MONS_SIXFIRHY },
  {  1,  5,  514, FLAT, MONS_NEQOXEC },
  {  1,  5,  514, FLAT, MONS_ORANGE_DEMON },
  {  1,  5,  514, FLAT, MONS_SMOKE_DEMON },
  {  1,  5,  514, FLAT, MONS_YNOXINUL },
  {  1,  5,   25, FLAT, MONS_CHAOS_SPAWN },
  {  1,  5,   25, FLAT, MONS_HELLION },
  {  1,  5,   25, FLAT, MONS_LOROCYPROCA },
  {  1,  5,   21, FLAT, MONS_TORMENTOR },
  {  1,  5,   17, FLAT, MONS_REAPER },
  {  1,  5,   14, FLAT, MONS_SOUL_EATER },
  {  1,  5,   25, FLAT, MONS_ICE_DEVIL },
  {  1,  5,   25, FLAT, MONS_BLUE_DEVIL },
  {  1,  5,   11, FLAT, MONS_IRON_DEVIL },
  {  1,  5,   25, FLAT, MONS_EXECUTIONER },
  {  1,  5,   25, FLAT, MONS_GREEN_DEATH },
  {  1,  5,   25, FLAT, MONS_BLIZZARD_DEMON },
  {  1,  5,   25, FLAT, MONS_BALRUG },
  {  1,  5,   25, FLAT, MONS_CACODEMON },
  {  1,  5,   25, FLAT, MONS_SUN_DEMON },
  {  1,  5,    8, FLAT, MONS_SHADOW_DEMON },
  {  1,  5,    8, FLAT, MONS_HELL_SENTINEL },
  {  1,  5,   73, FLAT, MONS_EFREET },
  {  1,  5,   66, FLAT, MONS_RAKSHASA },
  {  1,  5,   35, FLAT, MONS_UNSEEN_HORROR },
  {  1,  5,    8, FLAT, MONS_TENTACLED_MONSTROSITY },
  {  1,  5,  999, FLAT, MONS_ABOMINATION_SMALL },
  {  1,  5,  999, FLAT, MONS_ABOMINATION_LARGE },
  {  1,  5,   25, FLAT, MONS_NECROPHAGE },
  {  1,  5,   46, FLAT, MONS_FLAMING_CORPSE },
  {  1,  5,   25, FLAT, MONS_MUMMY },
  {  1,  5,   21, FLAT, MONS_VAMPIRE },
  {  1,  5,   17, FLAT, MONS_PHANTOM },
  {  1,  5,   30, FLAT, MONS_SHADOW },
  {  1,  5,   30, FLAT, MONS_HUNGRY_GHOST },
  {  1,  5,    6, FLAT, MONS_FLAYED_GHOST },
  {  1,  5,   25, FLAT, MONS_WIGHT },
  {  1,  5,   52, FLAT, MONS_WRAITH },
  {  1,  5,   52, FLAT, MONS_FREEZING_WRAITH },
  {  1,  5,    8, FLAT, MONS_SILENT_SPECTRE },
  {  1,  5,   17, FLAT, MONS_EIDOLON },
  {  1,  5,   52, FLAT, MONS_FLYING_SKULL },
  {  1,  5,    8, FLAT, MONS_PHANTASMAL_WARRIOR },
  {  1,  5,   46, FLAT, MONS_LICH },
  {  1,  5,    8, FLAT, MONS_ANCIENT_LICH },
  {  1,  5,   11, FLAT, MONS_PROFANE_SERVITOR },
  {  1,  5,  104, FLAT, MONS_SKELETON },
  {  1,  5,  104, FLAT, MONS_ZOMBIE },
  {  1,  5,   42, FLAT, MONS_SIMULACRUM },
  {  1,  5,   11, FLAT, MONS_ANGEL },
  {  1,  5,   14, FLAT, MONS_DAEVA },
  {  1,  5,  334, FLAT, MONS_WRETCHED_STAR },
  {  1,  5,  334, FLAT, MONS_TENTACLED_STARSPAWN },
  {  1,  5,  334, FLAT, MONS_ANCIENT_ZYME },
  {  1,  5,  334, FLAT, MONS_STARCURSED_MASS },
  {  1,  5,  334, FLAT, MONS_THRASHING_HORROR },
//{  1,  5,  334, FLAT, MONS_CHAOS_BUTTERFLY }, // whyyy disabled?
  {  1,  5,   89, FLAT, MONS_LURKING_HORROR },
  {  1,  5,   89, FLAT, MONS_SPATIAL_MAELSTROM },
  {  1,  5,   89, FLAT, MONS_APOCALYPSE_CRAB },
  { 0,0,0,FLAT,MONS_0 }
};

static const pop_entry pop_pan[] =
{ // Pandemonium
  {  1,  1,   25, FLAT, MONS_HELL_HOUND },
  {  1,  1,   25, FLAT, MONS_HELL_HOG },
  {  1,  1,   25, FLAT, MONS_HELLEPHANT },
  {  1,  1,    3, FLAT, MONS_GIANT_EYEBALL },
  {  1,  1,    3, FLAT, MONS_EYE_OF_DRAINING },
  {  1,  1,    3, FLAT, MONS_GREAT_ORB_OF_EYES },
  {  1,  1,    3, FLAT, MONS_GIANT_ORANGE_BRAIN },
  {  1,  1,    3, FLAT, MONS_GOLDEN_EYE },
  {  1,  1,  192, FLAT, MONS_DEMONIC_CRAWLER },
  {  1,  1,    8, FLAT, MONS_TOENAIL_GOLEM },
  {  1,  1,  334, FLAT, MONS_RED_DEVIL },
  {  1,  1,  334, FLAT, MONS_ROTTING_DEVIL },
  {  1,  1, 1000, FLAT, MONS_HELLWING },
  {  1,  1, 1000, FLAT, MONS_SIXFIRHY },
  {  1,  1, 1000, FLAT, MONS_NEQOXEC },
  {  1,  1, 1000, FLAT, MONS_ORANGE_DEMON },
  {  1,  1, 1000, FLAT, MONS_SMOKE_DEMON },
  {  1,  1, 1000, FLAT, MONS_YNOXINUL },
  {  1,  1,  192, FLAT, MONS_CHAOS_SPAWN },
  {  1,  1,  334, FLAT, MONS_LOROCYPROCA },
  {  1,  1,  334, FLAT, MONS_TORMENTOR },
  {  1,  1,  334, FLAT, MONS_REAPER },
  {  1,  1,  334, FLAT, MONS_SOUL_EATER },
  {  1,  1,  334, FLAT, MONS_ICE_DEVIL },
  {  1,  1,  334, FLAT, MONS_BLUE_DEVIL },
  {  1,  1,  334, FLAT, MONS_HELL_BEAST },
  {  1,  1,  334, FLAT, MONS_IRON_DEVIL },
  {  1,  1, 1000, FLAT, MONS_EXECUTIONER },
  {  1,  1, 1000, FLAT, MONS_GREEN_DEATH },
  {  1,  1, 1000, FLAT, MONS_BLIZZARD_DEMON },
  {  1,  1, 1000, FLAT, MONS_BALRUG },
  {  1,  1, 1000, FLAT, MONS_CACODEMON },
  {  1,  1,  334, FLAT, MONS_BRIMSTONE_FIEND },
  {  1,  1,  334, FLAT, MONS_ICE_FIEND },
  {  1,  1,  334, FLAT, MONS_SHADOW_FIEND },
  {  1,  1,  334, FLAT, MONS_HELL_SENTINEL },
  {  1,  1,  334, FLAT, MONS_SUN_DEMON },
  {  1,  1,  334, FLAT, MONS_SHADOW_DEMON },
  {  1,  1,  334, FLAT, MONS_HELLION },
  {  1,  1,   89, FLAT, MONS_EFREET },
  {  1,  1,   89, FLAT, MONS_RAKSHASA },
  {  1,  1, 1000, FLAT, MONS_ABOMINATION_LARGE },
  {  1,  1,    8, FLAT, MONS_PROFANE_SERVITOR },
  { 0,0,0,FLAT,MONS_0 }
};

static const pop_entry pop_zig[] =
{ // Ziggurat
  { 0,0,0,FLAT,MONS_0 }
};

static const pop_entry pop_lab[] =
{ // Labyrinth
  { 0,0,0,FLAT,MONS_0 }
};

static const pop_entry pop_bazaar[] =
{ // Bazaar
  { 0,0,0,FLAT,MONS_0 }
};

static const pop_entry pop_trove[] =
{ // Trove
  { 0,0,0,FLAT,MONS_0 }
};

static const pop_entry pop_sewer[] =
{ // Sewer
  {  1,  1, 1000, FLAT, MONS_GIANT_NEWT },
  {  1,  1,  514, FLAT, MONS_GIANT_GECKO },
  {  1,  1, 1000, FLAT, MONS_BAT },
  {  1,  1,  514, FLAT, MONS_BALL_PYTHON },
  {  1,  1,  514, FLAT, MONS_ADDER },
  {  1,  1,  514, FLAT, MONS_WORM },
  {  1,  1,  514, FLAT, MONS_OOZE },
  {  1,  1,  514, FLAT, MONS_GIANT_COCKROACH },
  {  1,  1,  514, FLAT, MONS_GIANT_MITE },
  { 0,0,0,FLAT,MONS_0 }
};

static const pop_entry pop_ossuary[] =
{ // Ossuary
  {  1,  1,   89, FLAT, MONS_MUMMY },
  {  1,  1,  514, FLAT, MONS_SKELETON },
  {  1,  1,  514, FLAT, MONS_ZOMBIE },
  { 0,0,0,FLAT,MONS_0 }
};

static const pop_entry pop_bailey[] =
{ // Bailey
  {  1,  1,  514, FLAT, MONS_GNOLL },
  {  1,  1,  514, FLAT, MONS_ORC },
  {  1,  1,  258, FLAT, MONS_ORC_WARRIOR },
  {  1,  1,   25, FLAT, MONS_ORC_KNIGHT },
  { 0,0,0,FLAT,MONS_0 }
};

static const pop_entry pop_icecv[] =
{ // Ice Cave
  {  1,  1,   89, FLAT, MONS_YAK },
  {  1,  1,   89, FLAT, MONS_POLAR_BEAR },
  {  1,  1,  514, FLAT, MONS_ICE_BEAST },
  {  1,  1,  514, FLAT, MONS_WHITE_IMP },
  {  1,  1,  514, FLAT, MONS_ICE_DEVIL },
  {  1,  1,   25, FLAT, MONS_BLUE_DEVIL },
  {  1,  1,  258, FLAT, MONS_FREEZING_WRAITH },
  {  1,  1, 1028, FLAT, MONS_SIMULACRUM },
  { 0,0,0,FLAT,MONS_0 }
};

static const pop_entry pop_volcano[] =
{ // Volcano
  {  1,  1,  514, FLAT, MONS_HELL_HOUND },
  {  1,  1,  514, FLAT, MONS_HELL_HOG },
  {  1,  1,  514, FLAT, MONS_MANTICORE },
  {  1,  1,  514, FLAT, MONS_LINDWURM },
  {  1,  1,  514, FLAT, MONS_FIRE_DRAKE },
  {  1,  1,  514, FLAT, MONS_TOENAIL_GOLEM },
  {  1,  1,  514, FLAT, MONS_FIRE_ELEMENTAL },
  {  1,  1,  514, FLAT, MONS_FIRE_VORTEX },
  {  1,  1,  514, FLAT, MONS_CRIMSON_IMP },
  {  1,  1,  514, FLAT, MONS_EFREET },
  {  1,  1,  514, FLAT, MONS_FLAYED_GHOST },
  { 0,0,0,FLAT,MONS_0 }
};

static const pop_entry pop_wizlab[] =
{ // Wizlab
  { 0,0,0,FLAT,MONS_0 }
};

#if TAG_MAJOR_VERSION == 34
static const pop_entry pop_removed[] =
{
  { 0,0,0,FLAT,MONS_0 }
};
#endif

// In C++ prior to C++11, we can't have an array of pointers to initialized
// arrays of different sizes without declaring individual arrays as
// standalone symbols first.  Thus this fugly list.
// Statically defined counts are merely an inconsequential nicety, no need
// to preserve them.
#define POP(x) { pop_##x, ARRAYSZ(pop_##x) - 1 }
static const struct { const pop_entry *pop; int count; } population[] =
{
    POP(d),
    POP(temple),
    POP(orc),
    POP(elf),
#if TAG_MAJOR_VERSION == 34
    POP(dwarf),
#endif
    POP(lair),
    POP(swamp),
    POP(shoals),
    POP(snake),
    POP(spider),
    POP(slime),
    POP(vaults),
    POP(blade),
    POP(crypt),
    POP(tomb),
    POP(hell),
    POP(dis),
    POP(geh),
    POP(coc),
    POP(tar),
    POP(zot),
    POP(forest),
    POP(abyss),
    POP(pan),
    POP(zig),
    POP(lab),
    POP(bazaar),
    POP(trove),
    POP(sewer),
    POP(ossuary),
    POP(bailey),
    POP(icecv),
    POP(volcano),
    POP(wizlab),
#if TAG_MAJOR_VERSION == 34
    POP(removed),
#endif
};

static const pop_entry pop_water_d[] =
{ // Dungeon water monsters
  {  5,  22,   80, FLAT, MONS_BIG_FISH },
  {  5,  24,  100, FLAT, MONS_GIANT_GOLDFISH },
  {  5,  16,   60, FLAT, MONS_ELECTRIC_EEL },
  {  7,  16,  185, PEAK, MONS_ELECTRIC_EEL },
  {  5,  20,  100, FLAT, MONS_JELLYFISH },
  {  9,  32,  200, SEMI, MONS_SHARK },
  {  11, 27,  600,   UP, MONS_WATER_ELEMENTAL },
  { 0,0,0,FLAT,MONS_0 }
};

static const pop_entry pop_water_forest[] =
{ // Forest water monsters
  {  1,   5,   35, FLAT, MONS_BIG_FISH },
  {  1,   5,  110, FLAT, MONS_GIANT_GOLDFISH },
  {  1,   5,   60, FLAT, MONS_JELLYFISH },
  {  1,   5,  125, FLAT, MONS_WATER_ELEMENTAL },
  { -1,   5,   65,   UP, MONS_ELEMENTAL_WELLSPRING },
  {  1,   5,  135, FLAT, MONS_THORN_LOTUS },
  {  1,   9,  120, SEMI, MONS_WATER_NYMPH },
  { 0,0,0,FLAT,MONS_0 }
};
